@@ -57,6 +57,7 @@ that the block has been replaced with the real block.
 @return           w->state() */
 inline uint32_t buf_pool_t::watch_remove(buf_page_t *w,
                                          buf_pool_t::hash_chain &chain)
+  noexcept
 {
   mysql_mutex_assert_owner(&buf_pool.mutex);
   ut_ad(xtest() || page_hash.lock_get(chain).is_write_locked());
@@ -95,7 +96,7 @@ and the lock released later.
 @retval	NULL	in case of an error */
 TRANSACTIONAL_TARGET
 static buf_page_t* buf_page_init_for_read(ulint mode, const page_id_t page_id,
-                                          ulint zip_size, bool unzip)
+                                          ulint zip_size, bool unzip) noexcept
 {
   mtr_t mtr;
 
@@ -266,7 +267,7 @@ buf_read_page_low(
 	ulint			mode,
 	const page_id_t		page_id,
 	ulint			zip_size,
-	bool			unzip)
+	bool			unzip) noexcept
 {
 	buf_page_t*	bpage;
 
@@ -357,7 +358,7 @@ wants to access
 pages, it may happen that the page at the given page number does not
 get read even if we return a positive value! */
 TRANSACTIONAL_TARGET
-ulint buf_read_ahead_random(const page_id_t page_id, bool ibuf)
+ulint buf_read_ahead_random(const page_id_t page_id, bool ibuf) noexcept
 {
   if (!srv_random_read_ahead || page_id.space() >= SRV_TMP_SPACE_ID)
     /* Disable the read-ahead for temporary tablespace */
@@ -447,7 +448,7 @@ read_ahead:
   return count;
 }
 
-dberr_t buf_read_page(const page_id_t page_id)
+dberr_t buf_read_page(const page_id_t page_id, bool unzip) noexcept
 {
   fil_space_t *space= fil_space_t::get(page_id.space());
   if (UNIV_UNLIKELY(!space))
@@ -462,7 +463,7 @@ dberr_t buf_read_page(const page_id_t page_id)
 
   buf_LRU_stat_inc_io(); /* NOT protected by buf_pool.mutex */
   return buf_read_page_low(space, true, BUF_READ_ANY_PAGE,
-                           page_id, space->zip_size(), true);
+                           page_id, space->zip_size(), unzip);
 }
 
 /** High-level function which reads a page asynchronously from a file to the
@@ -473,7 +474,7 @@ released by the i/o-handler thread.
 @param[in]	page_id		page id
 @param[in]	zip_size	ROW_FORMAT=COMPRESSED page size, or 0 */
 void buf_read_page_background(fil_space_t *space, const page_id_t page_id,
-                              ulint zip_size)
+                              ulint zip_size) noexcept
 {
 	buf_read_page_low(space, false, BUF_READ_ANY_PAGE,
 			  page_id, zip_size, false);
@@ -512,7 +513,7 @@ which could result in a deadlock if the OS does not support asynchronous io.
 @param[in]	ibuf		whether if we are inside ibuf routine
 @return number of page read requests issued */
 TRANSACTIONAL_TARGET
-ulint buf_read_ahead_linear(const page_id_t page_id, bool ibuf)
+ulint buf_read_ahead_linear(const page_id_t page_id, bool ibuf) noexcept
 {
   /* check if readahead is disabled.
   Disable the read ahead logic for temporary tablespace */
@@ -689,7 +690,7 @@ failed:
 @param recs     log records
 @param init     page initialization, or nullptr if the page needs to be read */
 void buf_read_recover(fil_space_t *space, const page_id_t page_id,
-                      page_recv_t &recs, recv_init *init)
+                      page_recv_t &recs, recv_init *init) noexcept
 {
   ut_ad(space->id == page_id.space());
   space->reacquire();
