@@ -1566,6 +1566,16 @@ group_by_handler *spider_create_group_by_handler(
         DBUG_PRINT("info",("spider select item=%p", item));
         if (item->const_item())
         {
+          /*
+            Do not create the GBH when a derived table or view is
+            involved
+          */
+          if (thd->derived_tables != NULL)
+          {
+            keep_going= FALSE;
+            break;
+          }
+
           /*
             Do not handle the complex case where there's a const item
             in the auxiliary fields. It is too unlikely (if at all) to
